const Queue = @import("./Queue.zig");

const coral = @import("coral");

const ext = @import("../ext.zig");

const gfx = @import("../gfx.zig");

const sokol = @import("sokol");

const std = @import("std");

thread: std.Thread,
clear_color: gfx.Color = gfx.colors.black,
state: *State,

const AtomicBool = std.atomic.Value(bool);

const Frame = struct {
	width: u16,
	height: u16,
	flushed_instance_2d_count: usize = 0,
	pushed_instance_2d_count: usize = 0,
	mesh_2d: gfx.Handle = .none,
	texture: gfx.Handle = .none,

	fn unflushed_instance_2d_count(self: Frame) usize {
		return self.pushed_instance_2d_count - self.flushed_instance_2d_count;
	}
};

const Render = struct {
	resources: coral.stack.Sequential(Resource),
	instance_2d_pipeline: sokol.gfx.Pipeline,
	instance_2d_buffers: coral.stack.Sequential(sokol.gfx.Buffer),

	const Instance2D = extern struct {
		transform: gfx.Transform2D,
		tint: @Vector(4, u8) = @splat(std.math.maxInt(u8)),
		depth: f32 = 0,
		texture_offset: gfx.Point2D = @splat(0),
		texture_size: gfx.Point2D = @splat(1),

		const buffer_indices = .{
			.mesh = 0,
			.instance = 1,
		};

		const instances_per_buffer = 512;

		const shader = @import("./shaders/instance_2d.glsl.zig");
	};

	const Resource = union (enum) {
		empty,
		mesh_2d: Mesh2D,
		texture: Texture,

		const Mesh2D = struct {
			index_count: u32,
			vertex_buffer: sokol.gfx.Buffer,
			index_buffer: sokol.gfx.Buffer,
		};

		const Texture = struct {
			image: sokol.gfx.Image,
			sampler: sokol.gfx.Sampler,
		};
	};

	fn deinit(self: *Render) void {
		for (self.instance_2d_buffers.values) |buffer| {
			sokol.gfx.destroyBuffer(buffer);
		}

		self.instance_2d_buffers.deinit();
		sokol.gfx.destroyPipeline(self.instance_2d_pipeline);
		self.resources.deinit();
	}

	fn init() Render {
		sokol.gfx.setup(.{
			.environment = .{
				.defaults = .{
					.color_format = .RGBA8,
					.depth_format = .DEPTH_STENCIL,
					.sample_count = 1,
				},
			},

			.logger = .{
				.func = sokol.log.func,
			},
		});

		return .{
			.instance_2d_pipeline = sokol.gfx.makePipeline(.{
				.label = "2D drawing pipeline",

				.layout = .{
					.attrs = get: {
						var attrs = [_]sokol.gfx.VertexAttrState{.{}} ** 16;

						attrs[Instance2D.shader.ATTR_vs_mesh_xy] = .{
							.format = .FLOAT2,
							.buffer_index = Instance2D.buffer_indices.mesh,
						};

						attrs[Instance2D.shader.ATTR_vs_mesh_uv] = .{
							.format = .FLOAT2,
							.buffer_index = Instance2D.buffer_indices.mesh,
						};

						attrs[Instance2D.shader.ATTR_vs_instance_xbasis] = .{
							.format = .FLOAT2,
							.buffer_index = Instance2D.buffer_indices.instance,
						};

						attrs[Instance2D.shader.ATTR_vs_instance_ybasis] = .{
							.format = .FLOAT2,
							.buffer_index = Instance2D.buffer_indices.instance,
						};

						attrs[Instance2D.shader.ATTR_vs_instance_origin] = .{
							.format = .FLOAT2,
							.buffer_index = Instance2D.buffer_indices.instance,
						};

						attrs[Instance2D.shader.ATTR_vs_instance_tint] = .{
							.format = .UBYTE4N,
							.buffer_index = Instance2D.buffer_indices.instance,
						};

						attrs[Instance2D.shader.ATTR_vs_instance_depth] = .{
							.format = .FLOAT,
							.buffer_index = Instance2D.buffer_indices.instance,
						};

						attrs[Instance2D.shader.ATTR_vs_instance_rect] = .{
							.format = .FLOAT4,
							.buffer_index = Instance2D.buffer_indices.instance,
						};

						break: get attrs;
					},

					.buffers = get: {
						var buffers = [_]sokol.gfx.VertexBufferLayoutState{.{}} ** 8;

						buffers[Instance2D.buffer_indices.instance].step_func = .PER_INSTANCE;

						break: get buffers;
					},
				},

				.shader = sokol.gfx.makeShader(Instance2D.shader.draw2dShaderDesc(sokol.gfx.queryBackend())),
				.index_type = .UINT16,
			}),

			.instance_2d_buffers = .{.allocator = coral.heap.allocator},
			.resources = .{.allocator = coral.heap.allocator},
		};
	}

	fn insert_resource(self: *Render, handle: gfx.Handle, resource: Resource) !void {
		const handle_index = handle.index() orelse {
			return error.InvalidHandle;
		};

		const resource_count = self.resources.len();

		if (handle_index < resource_count) {
			const empty_resource = &self.resources.values[handle_index];

			if (empty_resource.* != .empty) {
				return error.InvalidHandle;
			}

			empty_resource.* = resource;
		} else {
			if (handle_index != resource_count) {
				return error.InvalidIndex;
			}

			try self.resources.push(resource);
		}
	}

	fn flush_instance_2ds(self: *Render, frame: *Frame) void {
		const unflushed_count = frame.unflushed_instance_2d_count();

		if (unflushed_count == 0) {
			return;
		}

		sokol.gfx.applyPipeline(self.instance_2d_pipeline);

		sokol.gfx.applyUniforms(.VS, Instance2D.shader.SLOT_Screen, sokol.gfx.asRange(&Instance2D.shader.Screen{
			.screen_size = .{@floatFromInt(frame.width), @floatFromInt(frame.height)},
		}));

		const mesh_2d = self.resources.values[frame.mesh_2d.index().?].mesh_2d;
		const texture = self.resources.values[frame.texture.index().?].texture;

		var bindings = sokol.gfx.Bindings{
			.vertex_buffers = get: {
				var buffers = [_]sokol.gfx.Buffer{.{}} ** 8;

				buffers[Instance2D.buffer_indices.mesh] = mesh_2d.vertex_buffer;

				break: get buffers;
			},

			.index_buffer = mesh_2d.index_buffer,

			.fs = .{
				.images = get: {
					var images = [_]sokol.gfx.Image{.{}} ** 12;

					images[0] = texture.image;

					break: get images;
				},

				.samplers = get: {
					var samplers = [_]sokol.gfx.Sampler{.{}} ** 8;

					samplers[0] = texture.sampler;

					break: get samplers;
				},
			},
		};

		while (frame.flushed_instance_2d_count < frame.pushed_instance_2d_count) {
			const buffer_index = frame.flushed_instance_2d_count / Instance2D.instances_per_buffer;
			const buffer_offset = frame.flushed_instance_2d_count % Instance2D.instances_per_buffer;
			const instances_to_flush = @min(Instance2D.instances_per_buffer - buffer_offset, unflushed_count);

			bindings.vertex_buffers[Instance2D.buffer_indices.instance] = self.instance_2d_buffers.values[buffer_index];
			bindings.vertex_buffer_offsets[Instance2D.buffer_indices.instance] = @intCast(buffer_offset);

			sokol.gfx.applyBindings(bindings);
			sokol.gfx.draw(0, mesh_2d.index_count, @intCast(instances_to_flush));

			frame.flushed_instance_2d_count += instances_to_flush;
		}
	}

	fn push_instance_2d(self: *Render, frame: *Frame, command: Queue.DrawCommand.Instance) std.mem.Allocator.Error!void {
		if (command.mesh_2d != frame.mesh_2d or command.texture != frame.texture) {
			self.flush_instance_2ds(frame);
		}

		frame.mesh_2d = command.mesh_2d;
		frame.texture = command.texture;

		const has_filled_buffer = (frame.pushed_instance_2d_count % Instance2D.instances_per_buffer) == 0;
		const pushed_buffer_count = frame.pushed_instance_2d_count / Instance2D.instances_per_buffer;

		if (has_filled_buffer and pushed_buffer_count == self.instance_2d_buffers.len()) {
			const instance_buffer = sokol.gfx.makeBuffer(.{
				.size = @sizeOf(Instance2D) * Instance2D.instances_per_buffer,
				.usage = .STREAM,
				.label = "2D drawing instance buffer",
			});

			errdefer sokol.gfx.destroyBuffer(instance_buffer);

			try self.instance_2d_buffers.push(instance_buffer);
		}

		_ = sokol.gfx.appendBuffer(self.instance_2d_buffers.get().?, sokol.gfx.asRange(&Instance2D{
			.transform = command.transform,
		}));

		frame.pushed_instance_2d_count += 1;
	}

	fn remove_resource(self: *Render, handle: gfx.Handle) ?Resource {
		if (handle.index()) |handle_index| {
			const resource = self.resources.values[handle_index];

			if (resource != .empty) {
				self.resources.values[handle_index] = .empty;

				return resource;
			}
		}

		return null;
	}
};

const Self = @This();

const State = struct {
	finished: std.Thread.Semaphore = .{},
	is_running: AtomicBool = AtomicBool.init(true),
	ready: std.Thread.Semaphore = .{},
	clear_color: gfx.Color = gfx.colors.black,
};

pub fn deinit(self: *Self) void {
	self.state.is_running.store(false, .monotonic);
	self.state.ready.post();
	self.thread.join();
	coral.heap.allocator.destroy(self.state);

	self.* = undefined;
}

pub fn init(window: *ext.SDL_Window) (std.mem.Allocator.Error || std.Thread.SpawnError)!Self {
	const state = try coral.heap.allocator.create(State);

	errdefer coral.heap.allocator.destroy(state);

	state.* = .{};

	const thread = try std.Thread.spawn(.{}, run, .{window, state});

	thread.setName("Ona Graphics") catch {
		std.log.warn("failed to name the graphics thread", .{});
	};

	return .{
		.thread = thread,
		.state = state,
	};
}

fn process_close_command(command: Queue.CloseCommand, rendering: *Render) !void {
	const resource = &rendering.resources.values[command.handle.index().?];

	switch (resource.*) {
		.empty => {}, // TODO: Handle this.

		.mesh_2d => |mesh_2d| {
			sokol.gfx.destroyBuffer(mesh_2d.vertex_buffer);
			sokol.gfx.destroyBuffer(mesh_2d.index_buffer);
		},

		.texture => |texture| {
			sokol.gfx.destroyImage(texture.image);
			sokol.gfx.destroySampler(texture.sampler);
		},
	}

	resource.* = .empty;
}

fn process_draw_command(command: Queue.DrawCommand, render: *Render, frame: *Frame) !void {
	switch (command) {
		.instance_2d => |instance_2d| {
			try render.push_instance_2d(frame, instance_2d);
		},

		.post_process => |post_process| {
			render.flush_instance_2ds(frame);
			// sokol.gfx.applyPipeline(self.post_process_pipeline);

			_ = post_process;
		},
	}

	render.flush_instance_2ds(frame);
}

fn process_open_command(command: Queue.OpenCommand, render: *Render) !void {
	switch (command.resource) {
		.texture => |texture| {
			const stride = texture.width * texture.format.byte_size();

			const image = sokol.gfx.makeImage(.{
				.width = texture.width,
				.height = @intCast(texture.data.len / stride),

				.data = .{
					.subimage = get: {
						var subimage = [_][16]sokol.gfx.Range{.{.{}} ** 16} ** 6;

						subimage[0][0] = sokol.gfx.asRange(texture.data);

						break: get subimage;
					},
				},
			});

			errdefer sokol.gfx.destroyImage(image);

			const sampler = sokol.gfx.makeSampler(.{});

			errdefer sokol.gfx.destroySampler(sampler);

			try render.insert_resource(command.handle, .{
				.texture = .{
					.sampler = sampler,
					.image = image,
				},
			});
		},

		.mesh_2d => |mesh_2d| {
			const index_buffer = sokol.gfx.makeBuffer(.{
				.data = sokol.gfx.asRange(mesh_2d.indices),
				.type = .INDEXBUFFER,
			});

			const vertex_buffer = sokol.gfx.makeBuffer(.{
				.data = sokol.gfx.asRange(mesh_2d.vertices),
				.type = .VERTEXBUFFER,
			});

			errdefer {
				sokol.gfx.destroyBuffer(index_buffer);
				sokol.gfx.destroyBuffer(vertex_buffer);
			}

			if (mesh_2d.indices.len > std.math.maxInt(u32)) {
				return error.OutOfMemory;
			}

			try render.insert_resource(command.handle, .{
				.mesh_2d = .{
					.index_buffer = index_buffer,
					.vertex_buffer = vertex_buffer,
					.index_count = @intCast(mesh_2d.indices.len),
				},
			});
		},
	}
}

fn run(window: *ext.SDL_Window, state: *State) !void {
	const context = configure_and_create: {
		var result = @as(c_int, 0);

		result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_FLAGS, ext.SDL_GL_CONTEXT_FORWARD_COMPATIBLE_FLAG);
		result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_PROFILE_MASK, ext.SDL_GL_CONTEXT_PROFILE_CORE);
		result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_MAJOR_VERSION, 3);
		result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_MINOR_VERSION, 3);
		result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_DOUBLEBUFFER, 1);

		if (result != 0) {
			return error.Unsupported;
		}

		break: configure_and_create ext.SDL_GL_CreateContext(window);
	};

	sokol.gfx.setup(.{
		.environment = .{
			.defaults = .{
				.color_format = .RGBA8,
				.depth_format = .DEPTH_STENCIL,
				.sample_count = 1,
			},
		},

		.logger = .{
			.func = sokol.log.func,
		},
	});

	defer {
		sokol.gfx.shutdown();
		ext.SDL_GL_DeleteContext(context);
	}

	var render = Render.init();

	defer render.deinit();

	state.finished.post();

	while (state.is_running.load(.monotonic)) {
		state.ready.wait();

		defer state.finished.post();

		var frame = init_frame: {
			var width, var height = [_]c_int{0, 0};

			ext.SDL_GL_GetDrawableSize(window, &width, &height);
			std.debug.assert(width > 0 and height > 0);

			break: init_frame Frame{
				.width = @intCast(width),
				.height = @intCast(height),
			};
		};

		sokol.gfx.beginPass(.{
			.swapchain = .{
				.width = frame.width,
				.height = frame.height,
				.sample_count = 1,
				.color_format = .RGBA8,
				.depth_format = .DEPTH_STENCIL,
				.gl = .{.framebuffer = 0},
			},

			.action = .{
				.colors = get: {
					var actions = [_]sokol.gfx.ColorAttachmentAction{.{}} ** 4;

					actions[0] = .{
						.load_action = .CLEAR,
						.clear_value = @as(sokol.gfx.Color, @bitCast(state.clear_color)),
					};

					break: get actions;
				},
			},
		});

		try Queue.visit_open_commands(process_open_command, .{&render});
		try Queue.visit_draw_commands(process_draw_command, .{&render, &frame});
		try Queue.visit_close_commands(process_close_command, .{&render});

		sokol.gfx.endPass();
		sokol.gfx.commit();
		ext.SDL_GL_SwapWindow(window);
	}
}

pub fn submit(self: *Self) void {
	self.state.finished.wait();

	self.state.clear_color = self.clear_color;

	Queue.swap();
	self.state.ready.post();
}
